" Name:         murphy
" Description:  Green foreground black background.
" Author:       Original author Ron Aaron <ron@ronware.org>.
" Maintainer:   Original maintainer Ron Aaron <ron@ronware.org>.
" URL:          https://github.com/vim/colorschemes
" License:      Same as Vim
" Last Change:  2025 Oct 08

" Generated by Colortemplate v3.0.0-beta10

set background=dark

" hi clear
source $VIMRUNTIME/colors/vim.lua " Nvim: revert to Vim default color scheme
let g:colors_name = 'murphy'

let s:t_Co = &t_Co
let s:tgc = has('termguicolors') && &termguicolors

let g:terminal_ansi_colors = ['#303030', '#ffa700', '#005f00', '#ffd7af', '#87afff', '#ffafaf', '#00afaf', '#bcbcbc', '#444444', '#ff0000', '#00875f', '#ffff00', '#005fff', '#ff00ff', '#00ffff', '#ffffff']
" Nvim uses g:terminal_color_{0-15} instead
for i in range(g:terminal_ansi_colors->len())
  let g:terminal_color_{i} = g:terminal_ansi_colors[i]
endfor

hi! link Added Constant
hi! link CurSearch Search
hi! link CursorLineFold CursorLine
hi! link CursorLineSign CursorLine
hi! link Float Number
hi! link Function Identifier
hi! link LineNrAbove LineNr
hi! link LineNrBelow LineNr
hi! link MessageWindow Pmenu
hi! link Number Constant
hi! link PopupNotification Todo
hi! link StatusLineTerm StatusLine
hi! link StatusLineTermNC StatusLineNC
hi! link TabPanel Normal
hi! link TabPanelFill EndOfBuffer
hi! link Terminal Normal

hi Normal guifg=#87ff87 guibg=#000000 guisp=NONE gui=NONE ctermfg=120 ctermbg=16 cterm=NONE term=NONE
hi ColorColumn guifg=NONE guibg=#262626 guisp=NONE gui=NONE ctermfg=NONE ctermbg=235 cterm=NONE term=reverse
hi Comment guifg=#ffa700 guibg=NONE guisp=NONE gui=NONE ctermfg=214 ctermbg=NONE cterm=NONE term=bold
hi Conceal guifg=#bcbcbc guibg=NONE guisp=NONE gui=NONE ctermfg=250 ctermbg=NONE cterm=NONE term=NONE
hi Constant guifg=#ffffff guibg=NONE guisp=NONE gui=NONE ctermfg=231 ctermbg=NONE cterm=NONE term=NONE
hi Cursor guifg=#000000 guibg=#87ff87 guisp=NONE gui=NONE ctermfg=16 ctermbg=120 cterm=NONE term=reverse
hi CursorColumn guifg=NONE guibg=#444444 guisp=NONE gui=NONE ctermfg=NONE ctermbg=238 cterm=NONE term=NONE
hi CursorLine guifg=NONE guibg=#444444 guisp=NONE gui=NONE ctermfg=NONE ctermbg=238 cterm=NONE term=underline
hi CursorLineNr guifg=#ffff00 guibg=NONE guisp=NONE gui=bold ctermfg=226 ctermbg=NONE cterm=bold term=bold
hi DiffAdd guifg=#ffffff guibg=#5f875f guisp=NONE gui=NONE ctermfg=231 ctermbg=65 cterm=NONE term=reverse
hi DiffChange guifg=#ffffff guibg=#5f87af guisp=NONE gui=NONE ctermfg=231 ctermbg=67 cterm=NONE term=NONE
hi DiffDelete guifg=#ffffff guibg=#af5faf guisp=NONE gui=NONE ctermfg=231 ctermbg=133 cterm=NONE term=reverse
hi DiffText guifg=#000000 guibg=#c6c6c6 guisp=NONE gui=NONE ctermfg=16 ctermbg=251 cterm=NONE term=reverse
hi Directory guifg=#00ffff guibg=NONE guisp=NONE gui=NONE ctermfg=51 ctermbg=NONE cterm=NONE term=NONE
hi EndOfBuffer guifg=#0000ff guibg=#000000 guisp=NONE gui=NONE ctermfg=21 ctermbg=16 cterm=NONE term=NONE
hi Error guifg=#ffffff guibg=#ff0000 guisp=NONE gui=NONE ctermfg=231 ctermbg=196 cterm=NONE term=bold,reverse
hi ErrorMsg guifg=#ffffff guibg=#ff0000 guisp=NONE gui=NONE ctermfg=231 ctermbg=196 cterm=NONE term=bold,reverse
hi FoldColumn guifg=#00afaf guibg=NONE guisp=NONE gui=NONE ctermfg=37 ctermbg=NONE cterm=NONE term=NONE
hi Folded guifg=#00afaf guibg=#303030 guisp=NONE gui=NONE ctermfg=37 ctermbg=236 cterm=NONE term=NONE
hi Identifier guifg=#00ffff guibg=NONE guisp=NONE gui=NONE ctermfg=51 ctermbg=NONE cterm=NONE term=NONE
hi Ignore guifg=NONE guibg=NONE guisp=NONE gui=NONE ctermfg=NONE ctermbg=NONE cterm=NONE term=NONE
hi IncSearch guifg=#ffa700 guibg=NONE guisp=NONE gui=reverse ctermfg=214 ctermbg=NONE cterm=reverse term=bold,reverse,underline
hi LineNr guifg=#ffff00 guibg=NONE guisp=NONE gui=NONE ctermfg=226 ctermbg=NONE cterm=NONE term=NONE
hi MatchParen guifg=#303030 guibg=#ffd7af guisp=NONE gui=NONE ctermfg=236 ctermbg=223 cterm=NONE term=bold,underline
hi ModeMsg guifg=#87ff87 guibg=NONE guisp=NONE gui=bold ctermfg=120 ctermbg=NONE cterm=bold term=bold
hi MoreMsg guifg=#005f00 guibg=NONE guisp=NONE gui=bold ctermfg=22 ctermbg=NONE cterm=bold term=NONE
hi NonText guifg=#00afaf guibg=NONE guisp=NONE gui=NONE ctermfg=37 ctermbg=NONE cterm=NONE term=NONE
hi Pmenu guifg=#ffffff guibg=#444444 guisp=NONE gui=NONE ctermfg=231 ctermbg=238 cterm=NONE term=reverse
hi PmenuMatch guifg=#ff00ff guibg=#444444 guisp=NONE gui=NONE ctermfg=201 ctermbg=238 cterm=NONE term=NONE
hi PmenuMatchSel guifg=#ff00ff guibg=#ffff00 guisp=NONE gui=NONE ctermfg=201 ctermbg=226 cterm=NONE term=NONE
hi PmenuSbar guifg=NONE guibg=#303030 guisp=NONE gui=NONE ctermfg=NONE ctermbg=236 cterm=NONE term=reverse
hi PmenuSel guifg=#000000 guibg=#ffff00 guisp=NONE gui=NONE ctermfg=16 ctermbg=226 cterm=NONE term=bold
hi PmenuThumb guifg=NONE guibg=#bcbcbc guisp=NONE gui=NONE ctermfg=NONE ctermbg=250 cterm=NONE term=NONE
hi PreProc guifg=#ffd7af guibg=NONE guisp=NONE gui=NONE ctermfg=223 ctermbg=NONE cterm=NONE term=NONE
hi Question guifg=#00ffff guibg=NONE guisp=NONE gui=bold ctermfg=51 ctermbg=NONE cterm=bold term=standout
hi QuickFixLine guifg=#303030 guibg=#00afaf guisp=NONE gui=NONE ctermfg=236 ctermbg=37 cterm=NONE term=NONE
hi Search guifg=#ffffff guibg=#0000ff guisp=NONE gui=NONE ctermfg=231 ctermbg=21 cterm=NONE term=reverse
hi SignColumn guifg=#00afaf guibg=NONE guisp=NONE gui=NONE ctermfg=37 ctermbg=NONE cterm=NONE term=reverse
hi Special guifg=#ff00ff guibg=NONE guisp=NONE gui=NONE ctermfg=201 ctermbg=NONE cterm=NONE term=NONE
hi SpecialKey guifg=#00afaf guibg=NONE guisp=NONE gui=NONE ctermfg=37 ctermbg=NONE cterm=NONE term=bold
hi SpellBad guifg=#ff0000 guibg=NONE guisp=#ff0000 gui=undercurl ctermfg=196 ctermbg=NONE cterm=underline term=underline
hi SpellCap guifg=#00ffff guibg=NONE guisp=#00ffff gui=undercurl ctermfg=51 ctermbg=NONE cterm=underline term=underline
hi SpellLocal guifg=#ffd7af guibg=NONE guisp=#ffd7af gui=undercurl ctermfg=223 ctermbg=NONE cterm=underline term=underline
hi SpellRare guifg=#ffff00 guibg=NONE guisp=#ffff00 gui=undercurl ctermfg=226 ctermbg=NONE cterm=underline term=underline
hi Statement guifg=#ffff00 guibg=NONE guisp=NONE gui=NONE ctermfg=226 ctermbg=NONE cterm=NONE term=NONE
hi StatusLine guifg=#ffffff guibg=#00008b guisp=NONE gui=NONE ctermfg=231 ctermbg=18 cterm=NONE term=bold,reverse
hi StatusLineNC guifg=#ffffff guibg=#3a3a3a guisp=NONE gui=NONE ctermfg=231 ctermbg=237 cterm=NONE term=bold,underline
hi TabLine guifg=#87ff87 guibg=#444444 guisp=NONE gui=NONE ctermfg=120 ctermbg=238 cterm=NONE term=bold,underline
hi TabLineFill guifg=NONE guibg=#303030 guisp=NONE gui=NONE ctermfg=NONE ctermbg=236 cterm=NONE term=NONE
hi TabLineSel guifg=#ffffff guibg=#000000 guisp=NONE gui=NONE ctermfg=231 ctermbg=16 cterm=NONE term=bold,reverse
hi Title guifg=#ff00ff guibg=NONE guisp=NONE gui=bold ctermfg=201 ctermbg=NONE cterm=bold term=NONE
hi TitleBar guifg=#87ff87 guibg=#4c4c4c guisp=NONE gui=NONE ctermfg=120 ctermbg=237 cterm=NONE term=NONE
hi TitleBarNC guifg=#bcbcbc guibg=#3c3c3c guisp=NONE gui=NONE ctermfg=250 ctermbg=236 cterm=NONE term=NONE
hi Todo guifg=#0000ff guibg=#ffff00 guisp=NONE gui=NONE ctermfg=21 ctermbg=226 cterm=NONE term=bold,reverse
hi ToolbarButton guifg=#ffffff guibg=#444444 guisp=NONE gui=bold ctermfg=231 ctermbg=238 cterm=bold term=bold,reverse
hi ToolbarLine guifg=NONE guibg=#303030 guisp=NONE gui=NONE ctermfg=NONE ctermbg=236 cterm=NONE term=reverse
hi Type guifg=#bcbcbc guibg=NONE guisp=NONE gui=NONE ctermfg=250 ctermbg=NONE cterm=NONE term=NONE
hi Underlined guifg=#00afaf guibg=NONE guisp=NONE gui=underline ctermfg=37 ctermbg=NONE cterm=underline term=underline
hi VertSplit guifg=#ffffff guibg=#3a3a3a guisp=NONE gui=NONE ctermfg=231 ctermbg=237 cterm=NONE term=NONE
hi Visual guifg=#ffffff guibg=#005f00 guisp=NONE gui=NONE ctermfg=231 ctermbg=22 cterm=NONE term=reverse
hi VisualNOS guifg=#ffffff guibg=#005f00 guisp=NONE gui=NONE ctermfg=231 ctermbg=22 cterm=NONE term=NONE
hi WarningMsg guifg=#ffa700 guibg=NONE guisp=NONE gui=bold ctermfg=214 ctermbg=NONE cterm=bold term=standout
hi WildMenu guifg=#000000 guibg=#ffff00 guisp=NONE gui=NONE ctermfg=16 ctermbg=226 cterm=NONE term=bold
hi lCursor guifg=#000000 guibg=#ff0000 guisp=NONE gui=NONE ctermfg=16 ctermbg=196 cterm=NONE term=NONE

if s:tgc || s:t_Co >= 256
  finish
endif

if s:t_Co >= 16
  hi Normal ctermfg=green ctermbg=black cterm=NONE
  hi ColorColumn ctermfg=green ctermbg=darkgrey cterm=NONE
  hi Comment ctermfg=darkred ctermbg=NONE cterm=NONE
  hi Conceal ctermfg=grey ctermbg=NONE cterm=NONE
  hi Constant ctermfg=white ctermbg=NONE cterm=NONE
  hi Cursor ctermfg=black ctermbg=green cterm=NONE
  hi CursorColumn ctermfg=NONE ctermbg=NONE cterm=underline
  hi CursorLine ctermfg=NONE ctermbg=NONE cterm=underline
  hi CursorLineNr ctermfg=yellow ctermbg=NONE cterm=underline
  hi DiffAdd ctermfg=white ctermbg=darkgreen cterm=NONE
  hi DiffChange ctermfg=white ctermbg=blue cterm=NONE
  hi DiffDelete ctermfg=white ctermbg=magenta cterm=NONE
  hi DiffText ctermfg=black ctermbg=grey cterm=NONE
  hi Directory ctermfg=cyan ctermbg=NONE cterm=NONE
  hi EndOfBuffer ctermfg=blue ctermbg=black cterm=NONE
  hi Error ctermfg=white ctermbg=red cterm=NONE
  hi ErrorMsg ctermfg=white ctermbg=red cterm=NONE
  hi FoldColumn ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Folded ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Identifier ctermfg=cyan ctermbg=NONE cterm=NONE
  hi Ignore ctermfg=NONE ctermbg=NONE cterm=NONE
  hi IncSearch ctermfg=darkred ctermbg=NONE cterm=reverse
  hi LineNr ctermfg=yellow ctermbg=NONE cterm=NONE
  hi MatchParen ctermfg=black ctermbg=darkyellow cterm=NONE
  hi ModeMsg ctermfg=green ctermbg=NONE cterm=bold
  hi MoreMsg ctermfg=darkgreen ctermbg=NONE cterm=bold
  hi NonText ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Pmenu ctermfg=white ctermbg=darkgrey cterm=NONE
  hi PmenuMatch ctermfg=white ctermbg=darkgrey cterm=bold
  hi PmenuMatchSel ctermfg=black ctermbg=yellow cterm=bold
  hi PmenuSbar ctermfg=NONE ctermbg=black cterm=NONE
  hi PmenuSel ctermfg=black ctermbg=yellow cterm=NONE
  hi PmenuThumb ctermfg=NONE ctermbg=grey cterm=NONE
  hi PreProc ctermfg=darkmagenta ctermbg=NONE cterm=NONE
  hi Question ctermfg=cyan ctermbg=NONE cterm=bold
  hi QuickFixLine ctermfg=black ctermbg=darkcyan cterm=NONE
  hi Search ctermfg=white ctermbg=blue cterm=NONE
  hi SignColumn ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Special ctermfg=magenta ctermbg=NONE cterm=NONE
  hi SpecialKey ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi SpellBad ctermfg=red ctermbg=NONE cterm=underline
  hi SpellCap ctermfg=cyan ctermbg=NONE cterm=underline
  hi SpellLocal ctermfg=darkyellow ctermbg=NONE cterm=underline
  hi SpellRare ctermfg=yellow ctermbg=NONE cterm=underline
  hi Statement ctermfg=yellow ctermbg=NONE cterm=NONE
  hi StatusLine ctermfg=white ctermbg=darkblue cterm=NONE
  hi StatusLineNC ctermfg=white ctermbg=darkgrey cterm=NONE
  hi TabLine ctermfg=green ctermbg=darkgrey cterm=NONE
  hi TabLineFill ctermfg=NONE ctermbg=grey cterm=NONE
  hi TabLineSel ctermfg=white ctermbg=black cterm=NONE
  hi Title ctermfg=magenta ctermbg=NONE cterm=bold
  hi TitleBar ctermfg=green ctermbg=black cterm=NONE
  hi TitleBarNC ctermfg=grey ctermbg=black cterm=NONE
  hi Todo ctermfg=blue ctermbg=yellow cterm=NONE
  hi ToolbarButton ctermfg=white ctermbg=darkgrey cterm=bold
  hi ToolbarLine ctermfg=NONE ctermbg=black cterm=NONE
  hi Type ctermfg=grey ctermbg=NONE cterm=NONE
  hi Underlined ctermfg=darkcyan ctermbg=NONE cterm=underline
  hi VertSplit ctermfg=white ctermbg=darkgrey cterm=NONE
  hi Visual ctermfg=white ctermbg=darkgreen cterm=NONE
  hi VisualNOS ctermfg=white ctermbg=darkgreen cterm=NONE
  hi WarningMsg ctermfg=darkred ctermbg=NONE cterm=bold
  hi WildMenu ctermfg=black ctermbg=yellow cterm=NONE
  hi lCursor ctermfg=black ctermbg=red cterm=NONE
  finish
endif

if s:t_Co >= 8
  hi Normal ctermfg=darkgreen ctermbg=black cterm=NONE
  hi ColorColumn ctermfg=grey ctermbg=darkred cterm=NONE
  hi Comment ctermfg=darkred ctermbg=NONE cterm=bold
  hi Conceal ctermfg=grey ctermbg=NONE cterm=NONE
  hi Constant ctermfg=grey ctermbg=NONE cterm=bold
  hi Cursor ctermfg=black ctermbg=green cterm=NONE
  hi CursorColumn ctermfg=NONE ctermbg=NONE cterm=underline
  hi CursorLine ctermfg=NONE ctermbg=NONE cterm=underline
  hi CursorLineNr ctermfg=darkyellow ctermbg=NONE cterm=bold
  hi DiffAdd ctermfg=white ctermbg=darkgreen cterm=NONE
  hi DiffChange ctermfg=white ctermbg=darkblue cterm=NONE
  hi DiffDelete ctermfg=white ctermbg=darkmagenta cterm=NONE
  hi DiffText ctermfg=black ctermbg=grey cterm=NONE
  hi Directory ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi EndOfBuffer ctermfg=darkblue ctermbg=NONE cterm=bold
  hi Error ctermfg=grey ctermbg=darkred cterm=NONE
  hi ErrorMsg ctermfg=grey ctermbg=darkred cterm=NONE
  hi FoldColumn ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Folded ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Identifier ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Ignore ctermfg=NONE ctermbg=NONE cterm=NONE
  hi IncSearch ctermfg=darkred ctermbg=NONE cterm=reverse
  hi LineNr ctermfg=darkyellow ctermbg=NONE cterm=NONE
  hi MatchParen ctermfg=black ctermbg=darkyellow cterm=NONE
  hi ModeMsg ctermfg=grey ctermbg=NONE cterm=bold
  hi MoreMsg ctermfg=darkgreen ctermbg=NONE cterm=bold
  hi NonText ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Pmenu ctermfg=black ctermbg=darkcyan cterm=NONE
  hi PmenuMatch ctermfg=black ctermbg=darkcyan cterm=bold
  hi PmenuMatchSel ctermfg=black ctermbg=darkyellow cterm=bold
  hi PmenuSbar ctermfg=NONE ctermbg=black cterm=NONE
  hi PmenuSel ctermfg=black ctermbg=darkyellow cterm=NONE
  hi PmenuThumb ctermfg=NONE ctermbg=grey cterm=NONE
  hi PreProc ctermfg=darkmagenta ctermbg=NONE cterm=bold
  hi Question ctermfg=darkcyan ctermbg=NONE cterm=bold
  hi QuickFixLine ctermfg=black ctermbg=darkcyan cterm=NONE
  hi Search ctermfg=grey ctermbg=darkblue cterm=NONE
  hi SignColumn ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Special ctermfg=darkmagenta ctermbg=NONE cterm=NONE
  hi SpecialKey ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi SpellBad ctermfg=darkred ctermbg=darkyellow cterm=reverse
  hi SpellCap ctermfg=darkcyan ctermbg=NONE cterm=reverse
  hi SpellLocal ctermfg=darkyellow ctermbg=NONE cterm=reverse
  hi SpellRare ctermfg=darkmagenta ctermbg=darkyellow cterm=reverse
  hi Statement ctermfg=darkyellow ctermbg=NONE cterm=NONE
  hi StatusLine ctermfg=darkblue ctermbg=grey cterm=bold,reverse
  hi StatusLineNC ctermfg=grey ctermbg=black cterm=reverse
  hi TabLine ctermfg=grey ctermbg=black cterm=reverse
  hi TabLineFill ctermfg=NONE ctermbg=grey cterm=NONE
  hi TabLineSel ctermfg=grey ctermbg=black cterm=NONE
  hi Title ctermfg=darkmagenta ctermbg=NONE cterm=bold
  hi TitleBar ctermfg=green ctermbg=black cterm=NONE
  hi TitleBarNC ctermfg=grey ctermbg=black cterm=NONE
  hi Todo ctermfg=darkblue ctermbg=darkyellow cterm=NONE
  hi ToolbarButton ctermfg=grey ctermbg=black cterm=bold,reverse
  hi ToolbarLine ctermfg=NONE ctermbg=black cterm=NONE
  hi Type ctermfg=grey ctermbg=NONE cterm=NONE
  hi Underlined ctermfg=NONE ctermbg=NONE cterm=underline
  hi VertSplit ctermfg=grey ctermbg=black cterm=reverse
  hi Visual ctermfg=NONE ctermbg=NONE cterm=reverse
  hi VisualNOS ctermfg=NONE ctermbg=NONE cterm=reverse,underline
  hi WarningMsg ctermfg=darkred ctermbg=NONE cterm=bold
  hi WildMenu ctermfg=black ctermbg=darkyellow cterm=NONE
  hi lCursor ctermfg=black ctermbg=red cterm=NONE
  finish
endif

if s:t_Co >= 0
  hi CurSearch term=reverse
  hi CursorIM term=NONE
  hi CursorLineFold term=underline
  hi CursorLineSign term=underline
  hi Float term=NONE
  hi Function term=NONE
  hi Number term=NONE
  hi StatusLineTerm term=bold,reverse
  hi StatusLineTermNC term=bold,underline
  hi Terminal term=NONE
  finish
endif

" vim: et ts=8 sw=2 sts=2
