" Name:         blue
" Description:  Colorscheme with a blue background
" Author:       Original author Steven Vertigan <steven@vertigan.wattle.id.au>
" Maintainer:   Original maintainer Steven Vertigan <steven@vertigan.wattle.id.au>
" URL:          https://github.com/vim/colorschemes
" License:      Same as Vim
" Last Change:  2025 Oct 08

" Generated by Colortemplate v3.0.0-beta10

set background=dark

" hi clear
source $VIMRUNTIME/colors/vim.lua " Nvim: revert to Vim default color scheme
let g:colors_name = 'blue'

let s:t_Co = &t_Co
let s:tgc = has('termguicolors') && &termguicolors

let g:terminal_ansi_colors = ['#000000', '#cd0000', '#00cd00', '#cdcd00', '#0000ee', '#cd00cd', '#00cdcd', '#e5e5e5', '#7f7f7f', '#ff0000', '#00ff00', '#ffff00', '#5c5cff', '#ff00ff', '#00ffff', '#ffffff']
" Nvim uses g:terminal_color_{0-15} instead
for i in range(g:terminal_ansi_colors->len())
  let g:terminal_color_{i} = g:terminal_ansi_colors[i]
endfor

hi! link Added String
hi! link Boolean Constant
hi! link Character Constant
hi! link Conditional Statement
hi! link CurSearch Search
hi! link CursorLineFold CursorLine
hi! link CursorLineSign CursorLine
hi! link Debug Special
hi! link Define PreProc
hi! link Delimiter Special
hi! link Exception Statement
hi! link Float Constant
hi! link Function Identifier
hi! link Include PreProc
hi! link Keyword Statement
hi! link LineNrAbove LineNr
hi! link LineNrBelow LineNr
hi! link Macro PreProc
hi! link MessageWindow Pmenu
hi! link Number Constant
hi! link Operator Type
hi! link PopupNotification Todo
hi! link PreCondit PreProc
hi! link Removed WarningMsg
hi! link Repeat Statement
hi! link SpecialChar Special
hi! link SpecialComment Special
hi! link StatusLineTerm StatusLine
hi! link StatusLineTermNC StatusLineNC
hi! link StorageClass Type
hi! link String Constant
hi! link Structure Type
hi! link TabLine StatusLineNC
hi! link TabLineFill StatusLineNC
hi! link TabLineSel StatusLine
hi! link Tag Special
hi! link Terminal Normal
hi! link Typedef Type
hi! link diffBDiffer WarningMsg
hi! link diffCommon WarningMsg
hi! link diffDiffer WarningMsg
hi! link diffIdentical WarningMsg
hi! link diffIsA WarningMsg
hi! link diffNoEOL WarningMsg
hi! link diffOnly WarningMsg
hi! link lCursor Cursor

hi Normal guifg=#ffd700 guibg=#000087 guisp=NONE gui=NONE ctermfg=220 ctermbg=18 cterm=NONE term=NONE
hi ColorColumn guifg=NONE guibg=#870087 guisp=NONE gui=NONE ctermfg=NONE ctermbg=90 cterm=NONE term=reverse
hi Comment guifg=#878787 guibg=NONE guisp=NONE gui=bold ctermfg=102 ctermbg=NONE cterm=NONE term=bold
hi Conceal guifg=#008787 guibg=NONE guisp=NONE gui=NONE ctermfg=30 ctermbg=NONE cterm=NONE term=NONE
hi Constant guifg=#5fffff guibg=NONE guisp=NONE gui=NONE ctermfg=87 ctermbg=NONE cterm=NONE term=NONE
hi Cursor guifg=#000000 guibg=#00ff00 guisp=NONE gui=NONE ctermfg=16 ctermbg=46 cterm=NONE term=reverse
hi CursorColumn guifg=NONE guibg=#005faf guisp=NONE gui=NONE ctermfg=NONE ctermbg=25 cterm=NONE term=NONE
hi CursorIM guifg=#000000 guibg=#ffd700 guisp=NONE gui=NONE ctermfg=16 ctermbg=220 cterm=NONE term=NONE
hi CursorLine guifg=NONE guibg=#005faf guisp=NONE gui=NONE ctermfg=NONE ctermbg=25 cterm=NONE term=underline
hi CursorLineNr guifg=#ffd700 guibg=#005faf guisp=NONE gui=bold ctermfg=220 ctermbg=25 cterm=NONE term=bold
hi DiffAdd guifg=#ffffff guibg=#5f875f guisp=NONE gui=NONE ctermfg=231 ctermbg=65 cterm=NONE term=reverse
hi DiffChange guifg=#ffffff guibg=#5f87af guisp=NONE gui=NONE ctermfg=231 ctermbg=67 cterm=NONE term=NONE
hi DiffDelete guifg=#ffffff guibg=#af5faf guisp=NONE gui=NONE ctermfg=231 ctermbg=133 cterm=NONE term=reverse
hi DiffText guifg=#000000 guibg=#c6c6c6 guisp=NONE gui=NONE ctermfg=16 ctermbg=251 cterm=NONE term=reverse
hi Directory guifg=#5fffff guibg=NONE guisp=NONE gui=NONE ctermfg=87 ctermbg=NONE cterm=NONE term=NONE
hi EndOfBuffer guifg=#ffd700 guibg=#000087 guisp=NONE gui=NONE ctermfg=220 ctermbg=18 cterm=NONE term=NONE
hi Error guifg=#ff7f50 guibg=#000087 guisp=NONE gui=reverse ctermfg=209 ctermbg=18 cterm=reverse term=bold,reverse
hi ErrorMsg guifg=#ffffff guibg=#d70000 guisp=NONE gui=NONE ctermfg=231 ctermbg=160 cterm=NONE term=bold,reverse
hi FoldColumn guifg=#008787 guibg=NONE guisp=NONE gui=NONE ctermfg=30 ctermbg=NONE cterm=NONE term=NONE
hi Folded guifg=#000087 guibg=#878700 guisp=NONE gui=NONE ctermfg=18 ctermbg=100 cterm=NONE term=NONE
hi Identifier guifg=#bcbcbc guibg=NONE guisp=NONE gui=NONE ctermfg=250 ctermbg=NONE cterm=NONE term=NONE
hi Ignore guifg=#878787 guibg=NONE guisp=NONE gui=NONE ctermfg=102 ctermbg=NONE cterm=NONE term=NONE
hi IncSearch guifg=#d787d7 guibg=#000000 guisp=NONE gui=standout ctermfg=176 ctermbg=16 cterm=reverse term=bold,reverse,underline
hi Label guifg=#ffd700 guibg=NONE guisp=NONE gui=NONE ctermfg=220 ctermbg=NONE cterm=NONE term=NONE
hi LineNr guifg=#5fffff guibg=NONE guisp=NONE gui=NONE ctermfg=87 ctermbg=NONE cterm=NONE term=NONE
hi MatchParen guifg=NONE guibg=NONE guisp=NONE gui=reverse ctermfg=NONE ctermbg=NONE cterm=reverse term=bold,underline
hi ModeMsg guifg=#000087 guibg=#00ff00 guisp=NONE gui=NONE ctermfg=18 ctermbg=46 cterm=NONE term=bold
hi MoreMsg guifg=#5fffff guibg=NONE guisp=NONE gui=NONE ctermfg=87 ctermbg=NONE cterm=NONE term=NONE
hi NonText guifg=#d787d7 guibg=NONE guisp=NONE gui=NONE ctermfg=176 ctermbg=NONE cterm=NONE term=NONE
hi Pmenu guifg=#ffffff guibg=#008787 guisp=NONE gui=NONE ctermfg=231 ctermbg=30 cterm=NONE term=reverse
hi PmenuMatch guifg=#ffd700 guibg=#008787 guisp=NONE gui=NONE ctermfg=220 ctermbg=30 cterm=NONE term=NONE
hi PmenuMatchSel guifg=#ff7f50 guibg=#ffffff guisp=NONE gui=NONE ctermfg=209 ctermbg=231 cterm=NONE term=NONE
hi PmenuSbar guifg=NONE guibg=NONE guisp=NONE gui=NONE ctermfg=NONE ctermbg=NONE cterm=NONE term=reverse
hi PmenuSel guifg=#008787 guibg=#ffffff guisp=NONE gui=NONE ctermfg=30 ctermbg=231 cterm=NONE term=bold
hi PmenuThumb guifg=NONE guibg=#ffffff guisp=NONE gui=NONE ctermfg=NONE ctermbg=231 cterm=NONE term=NONE
hi PreProc guifg=#00ff00 guibg=NONE guisp=NONE gui=NONE ctermfg=46 ctermbg=NONE cterm=NONE term=NONE
hi Question guifg=#00ff00 guibg=NONE guisp=NONE gui=NONE ctermfg=46 ctermbg=NONE cterm=NONE term=standout
hi QuickFixLine guifg=#000000 guibg=#d787d7 guisp=NONE gui=NONE ctermfg=16 ctermbg=176 cterm=NONE term=NONE
hi Search guifg=#ffd700 guibg=#000000 guisp=NONE gui=reverse ctermfg=220 ctermbg=16 cterm=reverse term=reverse
hi SignColumn guifg=#008787 guibg=NONE guisp=NONE gui=NONE ctermfg=30 ctermbg=NONE cterm=NONE term=reverse
hi Special guifg=#d787d7 guibg=NONE guisp=NONE gui=NONE ctermfg=176 ctermbg=NONE cterm=NONE term=NONE
hi SpecialKey guifg=#5fffff guibg=NONE guisp=NONE gui=NONE ctermfg=87 ctermbg=NONE cterm=NONE term=bold
hi SpellBad guifg=#d70000 guibg=NONE guisp=#d70000 gui=undercurl ctermfg=160 ctermbg=NONE cterm=underline term=underline
hi SpellCap guifg=#00ff00 guibg=NONE guisp=#00ff00 gui=undercurl ctermfg=46 ctermbg=NONE cterm=underline term=underline
hi SpellLocal guifg=#ffffff guibg=NONE guisp=#ffffff gui=undercurl ctermfg=231 ctermbg=NONE cterm=underline term=underline
hi SpellRare guifg=#d787d7 guibg=NONE guisp=#d787d7 gui=undercurl ctermfg=176 ctermbg=NONE cterm=underline term=underline
hi Statement guifg=#ffffff guibg=NONE guisp=NONE gui=NONE ctermfg=231 ctermbg=NONE cterm=NONE term=NONE
hi StatusLine guifg=#000087 guibg=#5fffff guisp=NONE gui=NONE ctermfg=18 ctermbg=87 cterm=NONE term=bold,reverse
hi StatusLineNC guifg=#000087 guibg=#008787 guisp=NONE gui=NONE ctermfg=18 ctermbg=30 cterm=NONE term=bold,underline
hi Title guifg=#d787d7 guibg=NONE guisp=NONE gui=NONE ctermfg=176 ctermbg=NONE cterm=NONE term=NONE
hi TitleBar guifg=#ffd700 guibg=#1f4f87 guisp=NONE gui=NONE ctermfg=220 ctermbg=237 cterm=NONE term=NONE
hi TitleBarNC guifg=#878787 guibg=#0f3f77 guisp=NONE gui=NONE ctermfg=102 ctermbg=236 cterm=NONE term=NONE
hi Todo guifg=NONE guibg=NONE guisp=NONE gui=reverse ctermfg=NONE ctermbg=NONE cterm=reverse term=bold,reverse
hi ToolbarButton guifg=#ffffff guibg=#005faf guisp=NONE gui=NONE ctermfg=231 ctermbg=25 cterm=NONE term=bold,reverse
hi ToolbarLine guifg=NONE guibg=NONE guisp=NONE gui=NONE ctermfg=NONE ctermbg=NONE cterm=NONE term=reverse
hi Type guifg=#ffa500 guibg=NONE guisp=NONE gui=bold ctermfg=214 ctermbg=NONE cterm=NONE term=NONE
hi Underlined guifg=NONE guibg=NONE guisp=NONE gui=underline ctermfg=NONE ctermbg=NONE cterm=underline term=underline
hi VertSplit guifg=#008787 guibg=NONE guisp=NONE gui=NONE ctermfg=30 ctermbg=NONE cterm=NONE term=NONE
hi Visual guifg=#ffffff guibg=#008787 guisp=NONE gui=NONE ctermfg=231 ctermbg=30 cterm=NONE term=reverse
hi VisualNOS guifg=#008787 guibg=#ffffff guisp=NONE gui=NONE ctermfg=30 ctermbg=231 cterm=NONE term=NONE
hi WarningMsg guifg=#d787d7 guibg=NONE guisp=NONE gui=NONE ctermfg=176 ctermbg=NONE cterm=NONE term=standout
hi WildMenu guifg=#000087 guibg=#ffd700 guisp=NONE gui=NONE ctermfg=18 ctermbg=220 cterm=NONE term=bold
hi debugBreakpoint guifg=#00ff00 guibg=#000087 guisp=NONE gui=reverse ctermfg=46 ctermbg=18 cterm=reverse term=reverse
hi debugPC guifg=#5fffff guibg=#000087 guisp=NONE gui=reverse ctermfg=87 ctermbg=18 cterm=reverse term=reverse

if s:tgc || s:t_Co >= 256
  hi! link CursorLineNr CursorLine
  finish
endif

if s:t_Co >= 16
  hi! link CursorLineNr CursorLine
  hi Normal ctermfg=yellow ctermbg=darkblue cterm=NONE
  hi ColorColumn ctermfg=NONE ctermbg=darkmagenta cterm=NONE
  hi Comment ctermfg=darkgray ctermbg=NONE cterm=NONE
  hi Conceal ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Constant ctermfg=cyan ctermbg=NONE cterm=NONE
  hi Cursor ctermfg=black ctermbg=green cterm=NONE
  hi CursorColumn ctermfg=NONE ctermbg=blue cterm=NONE
  hi CursorIM ctermfg=black ctermbg=yellow cterm=NONE
  hi CursorLine ctermfg=NONE ctermbg=NONE cterm=underline
  hi DiffAdd ctermfg=white ctermbg=darkgreen cterm=NONE
  hi DiffChange ctermfg=white ctermbg=blue cterm=NONE
  hi DiffDelete ctermfg=white ctermbg=magenta cterm=NONE
  hi DiffText ctermfg=black ctermbg=grey cterm=NONE
  hi Directory ctermfg=cyan ctermbg=NONE cterm=NONE
  hi EndOfBuffer ctermfg=yellow ctermbg=darkblue cterm=NONE
  hi Error ctermfg=red ctermbg=darkblue cterm=reverse
  hi ErrorMsg ctermfg=white ctermbg=red cterm=NONE
  hi FoldColumn ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Folded ctermfg=darkblue ctermbg=darkyellow cterm=NONE
  hi Identifier ctermfg=gray ctermbg=NONE cterm=NONE
  hi Ignore ctermfg=darkgray ctermbg=NONE cterm=NONE
  hi IncSearch ctermfg=magenta ctermbg=black cterm=reverse
  hi Label ctermfg=yellow ctermbg=NONE cterm=NONE
  hi LineNr ctermfg=cyan ctermbg=NONE cterm=NONE
  hi MatchParen ctermfg=NONE ctermbg=NONE cterm=reverse
  hi ModeMsg ctermfg=darkblue ctermbg=green cterm=NONE
  hi MoreMsg ctermfg=cyan ctermbg=NONE cterm=NONE
  hi NonText ctermfg=magenta ctermbg=NONE cterm=NONE
  hi Pmenu ctermfg=black ctermbg=darkcyan cterm=NONE
  hi PmenuMatch ctermfg=black ctermbg=darkcyan cterm=bold
  hi PmenuMatchSel ctermfg=black ctermbg=white cterm=bold
  hi PmenuSbar ctermfg=NONE ctermbg=NONE cterm=NONE
  hi PmenuSel ctermfg=black ctermbg=white cterm=NONE
  hi PmenuThumb ctermfg=NONE ctermbg=white cterm=NONE
  hi PreProc ctermfg=green ctermbg=NONE cterm=NONE
  hi Question ctermfg=green ctermbg=NONE cterm=NONE
  hi QuickFixLine ctermfg=black ctermbg=magenta cterm=NONE
  hi Search ctermfg=yellow ctermbg=black cterm=reverse
  hi SignColumn ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Special ctermfg=magenta ctermbg=NONE cterm=NONE
  hi SpecialKey ctermfg=cyan ctermbg=NONE cterm=NONE
  hi SpellBad ctermfg=red ctermbg=NONE cterm=underline
  hi SpellCap ctermfg=green ctermbg=NONE cterm=underline
  hi SpellLocal ctermfg=white ctermbg=NONE cterm=underline
  hi SpellRare ctermfg=magenta ctermbg=NONE cterm=underline
  hi Statement ctermfg=white ctermbg=NONE cterm=NONE
  hi StatusLine ctermfg=darkblue ctermbg=cyan cterm=NONE
  hi StatusLineNC ctermfg=darkblue ctermbg=darkcyan cterm=NONE
  hi Title ctermfg=magenta ctermbg=NONE cterm=NONE
  hi TitleBar ctermfg=yellow ctermbg=black cterm=NONE
  hi TitleBarNC ctermfg=darkgray ctermbg=black cterm=NONE
  hi Todo ctermfg=NONE ctermbg=NONE cterm=reverse
  hi ToolbarButton ctermfg=white ctermbg=blue cterm=NONE
  hi ToolbarLine ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Type ctermfg=darkyellow ctermbg=NONE cterm=NONE
  hi Underlined ctermfg=NONE ctermbg=NONE cterm=underline
  hi VertSplit ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Visual ctermfg=white ctermbg=darkcyan cterm=NONE
  hi VisualNOS ctermfg=darkcyan ctermbg=white cterm=NONE
  hi WarningMsg ctermfg=magenta ctermbg=NONE cterm=NONE
  hi WildMenu ctermfg=darkblue ctermbg=yellow cterm=NONE
  hi debugBreakpoint ctermfg=green ctermbg=darkblue cterm=reverse
  hi debugPC ctermfg=cyan ctermbg=darkblue cterm=reverse
  finish
endif

if s:t_Co >= 8
  hi! link CursorLineNr CursorLine
  hi Normal ctermfg=darkyellow ctermbg=blue cterm=NONE
  hi ColorColumn ctermfg=NONE ctermbg=magenta cterm=NONE
  hi Comment ctermfg=black ctermbg=NONE cterm=bold
  hi Conceal ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Constant ctermfg=cyan ctermbg=NONE cterm=NONE
  hi Cursor ctermfg=black ctermbg=green cterm=NONE
  hi CursorColumn ctermfg=blue ctermbg=yellow cterm=bold,reverse
  hi CursorIM ctermfg=black ctermbg=yellow cterm=NONE
  hi CursorLine ctermfg=NONE ctermbg=NONE cterm=underline
  hi DiffAdd ctermfg=white ctermbg=darkgreen cterm=NONE
  hi DiffChange ctermfg=white ctermbg=darkblue cterm=NONE
  hi DiffDelete ctermfg=white ctermbg=darkmagenta cterm=NONE
  hi DiffText ctermfg=black ctermbg=grey cterm=NONE
  hi Directory ctermfg=cyan ctermbg=NONE cterm=NONE
  hi EndOfBuffer ctermfg=yellow ctermbg=blue cterm=NONE
  hi Error ctermfg=red ctermbg=blue cterm=reverse
  hi ErrorMsg ctermfg=gray ctermbg=red cterm=NONE
  hi FoldColumn ctermfg=cyan ctermbg=NONE cterm=NONE
  hi Folded ctermfg=blue ctermbg=yellow cterm=NONE
  hi Identifier ctermfg=gray ctermbg=NONE cterm=NONE
  hi Ignore ctermfg=gray ctermbg=NONE cterm=NONE
  hi IncSearch ctermfg=magenta ctermbg=black cterm=reverse
  hi Label ctermfg=yellow ctermbg=NONE cterm=NONE
  hi LineNr ctermfg=cyan ctermbg=NONE cterm=NONE
  hi MatchParen ctermfg=NONE ctermbg=NONE cterm=reverse
  hi ModeMsg ctermfg=blue ctermbg=green cterm=NONE
  hi MoreMsg ctermfg=cyan ctermbg=NONE cterm=NONE
  hi NonText ctermfg=magenta ctermbg=NONE cterm=NONE
  hi Pmenu ctermfg=black ctermbg=cyan cterm=NONE
  hi PmenuMatch ctermfg=black ctermbg=cyan cterm=bold
  hi PmenuMatchSel ctermfg=black ctermbg=gray cterm=bold
  hi PmenuSbar ctermfg=NONE ctermbg=NONE cterm=NONE
  hi PmenuSel ctermfg=black ctermbg=gray cterm=NONE
  hi PmenuThumb ctermfg=NONE ctermbg=white cterm=NONE
  hi PreProc ctermfg=green ctermbg=NONE cterm=NONE
  hi Question ctermfg=green ctermbg=NONE cterm=NONE
  hi QuickFixLine ctermfg=black ctermbg=magenta cterm=NONE
  hi Search ctermfg=yellow ctermbg=black cterm=reverse
  hi SignColumn ctermfg=cyan ctermbg=NONE cterm=NONE
  hi Special ctermfg=magenta ctermbg=NONE cterm=NONE
  hi SpecialKey ctermfg=cyan ctermbg=NONE cterm=NONE
  hi SpellBad ctermfg=red ctermbg=yellow cterm=reverse
  hi SpellCap ctermfg=green ctermbg=NONE cterm=reverse
  hi SpellLocal ctermfg=gray ctermbg=NONE cterm=reverse
  hi SpellRare ctermfg=magenta ctermbg=yellow cterm=reverse
  hi Statement ctermfg=gray ctermbg=NONE cterm=NONE
  hi StatusLine ctermfg=cyan ctermbg=blue cterm=bold,reverse
  hi StatusLineNC ctermfg=blue ctermbg=cyan cterm=NONE
  hi Title ctermfg=magenta ctermbg=NONE cterm=NONE
  hi TitleBar ctermfg=yellow ctermbg=black cterm=NONE
  hi TitleBarNC ctermfg=darkgray ctermbg=black cterm=NONE
  hi Todo ctermfg=NONE ctermbg=NONE cterm=reverse
  hi ToolbarButton ctermfg=blue ctermbg=gray cterm=bold,reverse
  hi ToolbarLine ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Type ctermfg=red ctermbg=NONE cterm=NONE
  hi Underlined ctermfg=NONE ctermbg=NONE cterm=underline
  hi VertSplit ctermfg=cyan ctermbg=NONE cterm=NONE
  hi Visual ctermfg=black ctermbg=cyan cterm=NONE
  hi VisualNOS ctermfg=cyan ctermbg=gray cterm=NONE
  hi WarningMsg ctermfg=red ctermbg=NONE cterm=NONE
  hi WildMenu ctermfg=blue ctermbg=yellow cterm=NONE
  hi debugBreakpoint ctermfg=green ctermbg=blue cterm=reverse
  hi debugPC ctermfg=cyan ctermbg=blue cterm=reverse
  finish
endif

if s:t_Co >= 0
  hi CurSearch term=reverse
  hi CursorLineFold term=underline
  hi CursorLineSign term=underline
  hi Float term=NONE
  hi Function term=NONE
  hi Number term=NONE
  hi StatusLineTerm term=bold,reverse
  hi StatusLineTermNC term=bold,underline
  hi TabLine term=bold,underline
  hi TabLineFill term=NONE
  hi TabLineSel term=bold,reverse
  hi Terminal term=NONE
  finish
endif

" vim: et ts=8 sw=2 sts=2
