" Name:         desert
" Description:  Light background colorscheme.
" Author:       Original author Hans Fugal <hans@fugal.net>
" Maintainer:   Original maintainer Hans Fugal <hans@fugal.net>
" URL:          https://github.com/vim/colorschemes
" License:      Same as Vim
" Last Change:  2025 Oct 08

" Generated by Colortemplate v3.0.0-beta10

set background=dark

" hi clear
source $VIMRUNTIME/colors/vim.lua " Nvim: revert to Vim default color scheme
let g:colors_name = 'desert'

let s:t_Co = &t_Co
let s:tgc = has('termguicolors') && &termguicolors

let g:terminal_ansi_colors = ['#7f7f8c', '#cd5c5c', '#9acd32', '#bdb76b', '#75a0ff', '#eeee00', '#cd853f', '#666666', '#8a7f7f', '#ff0000', '#89fb98', '#f0e68c', '#6dceeb', '#ffde9b', '#ffa0a0', '#c2bfa5']
" Nvim uses g:terminal_color_{0-15} instead
for i in range(g:terminal_ansi_colors->len())
  let g:terminal_color_{i} = g:terminal_ansi_colors[i]
endfor

hi! link CurSearch Search
hi! link CursorLineFold CursorLine
hi! link CursorLineSign CursorLine
hi! link EndOfBuffer NonText
hi! link Float Number
hi! link Function Identifier
hi! link LineNrAbove LineNr
hi! link LineNrBelow LineNr
hi! link MessageWindow Pmenu
hi! link Number Constant
hi! link PopupNotification Todo
hi! link TabPanel Normal
hi! link TabPanelFill EndOfBuffer
hi! link Terminal Normal

hi Normal guifg=#ffffff guibg=#333333 guisp=NONE gui=NONE ctermfg=231 ctermbg=236 cterm=NONE term=NONE
hi ColorColumn guifg=#ffffff guibg=#cd5c5c guisp=NONE gui=NONE ctermfg=231 ctermbg=167 cterm=NONE term=reverse
hi Comment guifg=#6dceeb guibg=NONE guisp=NONE gui=NONE ctermfg=81 ctermbg=NONE cterm=NONE term=bold
hi Conceal guifg=#666666 guibg=NONE guisp=NONE gui=NONE ctermfg=241 ctermbg=NONE cterm=NONE term=NONE
hi Constant guifg=#ffa0a0 guibg=NONE guisp=NONE gui=NONE ctermfg=217 ctermbg=NONE cterm=NONE term=NONE
hi CursorColumn guifg=NONE guibg=#666666 guisp=NONE gui=NONE ctermfg=NONE ctermbg=241 cterm=NONE term=NONE
hi CursorLine guifg=NONE guibg=#666666 guisp=NONE gui=NONE ctermfg=NONE ctermbg=241 cterm=NONE term=underline
hi CursorLineNr guifg=#eeee00 guibg=NONE guisp=NONE gui=bold ctermfg=226 ctermbg=NONE cterm=bold term=bold
hi DiffAdd guifg=#ffffff guibg=#5f875f guisp=NONE gui=NONE ctermfg=231 ctermbg=65 cterm=NONE term=reverse
hi DiffChange guifg=#ffffff guibg=#5f87af guisp=NONE gui=NONE ctermfg=231 ctermbg=67 cterm=NONE term=NONE
hi DiffDelete guifg=#ffffff guibg=#af5faf guisp=NONE gui=NONE ctermfg=231 ctermbg=133 cterm=NONE term=reverse
hi DiffText guifg=#000000 guibg=#c6c6c6 guisp=NONE gui=NONE ctermfg=16 ctermbg=251 cterm=NONE term=reverse
hi Directory guifg=#6dceeb guibg=NONE guisp=NONE gui=NONE ctermfg=81 ctermbg=NONE cterm=NONE term=NONE
hi Error guifg=#ff0000 guibg=#ffffff guisp=NONE gui=reverse ctermfg=196 ctermbg=231 cterm=reverse term=bold,reverse
hi ErrorMsg guifg=#ff0000 guibg=#ffffff guisp=NONE gui=reverse ctermfg=196 ctermbg=231 cterm=reverse term=bold,reverse
hi FoldColumn guifg=#eeee00 guibg=#4d4d4d guisp=NONE gui=NONE ctermfg=226 ctermbg=239 cterm=NONE term=NONE
hi Folded guifg=#eeee00 guibg=#4d4d4d guisp=NONE gui=NONE ctermfg=226 ctermbg=239 cterm=NONE term=NONE
hi Identifier guifg=#89fb98 guibg=NONE guisp=NONE gui=NONE ctermfg=120 ctermbg=NONE cterm=NONE term=NONE
hi Ignore guifg=NONE guibg=NONE guisp=NONE gui=NONE ctermfg=NONE ctermbg=NONE cterm=NONE term=NONE
hi IncSearch guifg=#f0e68c guibg=#cd853f guisp=NONE gui=NONE ctermfg=186 ctermbg=172 cterm=NONE term=bold,reverse,underline
hi LineNr guifg=#eeee00 guibg=NONE guisp=NONE gui=NONE ctermfg=226 ctermbg=NONE cterm=NONE term=NONE
hi MatchParen guifg=#7f7f8c guibg=#bdb76b guisp=NONE gui=NONE ctermfg=242 ctermbg=143 cterm=NONE term=bold,underline
hi ModeMsg guifg=#ffde9b guibg=NONE guisp=NONE gui=bold ctermfg=222 ctermbg=NONE cterm=bold term=bold
hi MoreMsg guifg=#9acd32 guibg=NONE guisp=NONE gui=bold ctermfg=112 ctermbg=NONE cterm=bold term=NONE
hi NonText guifg=#6dceeb guibg=#4d4d4d guisp=NONE gui=NONE ctermfg=81 ctermbg=239 cterm=NONE term=NONE
hi Pmenu guifg=#ffffff guibg=#666666 guisp=NONE gui=NONE ctermfg=231 ctermbg=241 cterm=NONE term=reverse
hi PmenuMatch guifg=#ffa0a0 guibg=#666666 guisp=NONE gui=NONE ctermfg=217 ctermbg=241 cterm=NONE term=NONE
hi PmenuMatchSel guifg=#cd5c5c guibg=#f0e68c guisp=NONE gui=NONE ctermfg=167 ctermbg=186 cterm=NONE term=NONE
hi PmenuSbar guifg=NONE guibg=#333333 guisp=NONE gui=NONE ctermfg=NONE ctermbg=236 cterm=NONE term=reverse
hi PmenuSel guifg=#333333 guibg=#f0e68c guisp=NONE gui=NONE ctermfg=236 ctermbg=186 cterm=NONE term=bold
hi PmenuThumb guifg=NONE guibg=#c2bfa5 guisp=NONE gui=NONE ctermfg=NONE ctermbg=144 cterm=NONE term=NONE
hi PreProc guifg=#cd5c5c guibg=NONE guisp=NONE gui=NONE ctermfg=167 ctermbg=NONE cterm=NONE term=NONE
hi Question guifg=#89fb98 guibg=NONE guisp=NONE gui=bold ctermfg=120 ctermbg=NONE cterm=bold term=standout
hi QuickFixLine guifg=#333333 guibg=#f0e68c guisp=NONE gui=NONE ctermfg=236 ctermbg=186 cterm=NONE term=NONE
hi Search guifg=#f0e68c guibg=#7f7f8c guisp=NONE gui=NONE ctermfg=186 ctermbg=242 cterm=NONE term=reverse
hi SignColumn guifg=NONE guibg=NONE guisp=NONE gui=NONE ctermfg=NONE ctermbg=NONE cterm=NONE term=reverse
hi Special guifg=#ffde9b guibg=NONE guisp=NONE gui=NONE ctermfg=222 ctermbg=NONE cterm=NONE term=NONE
hi SpecialKey guifg=#9acd32 guibg=NONE guisp=NONE gui=NONE ctermfg=112 ctermbg=NONE cterm=NONE term=bold
hi SpellBad guifg=#cd5c5c guibg=NONE guisp=#cd5c5c gui=undercurl ctermfg=167 ctermbg=NONE cterm=underline term=underline
hi SpellCap guifg=#75a0ff guibg=NONE guisp=#75a0ff gui=undercurl ctermfg=111 ctermbg=NONE cterm=underline term=underline
hi SpellLocal guifg=#ffde9b guibg=NONE guisp=#ffde9b gui=undercurl ctermfg=222 ctermbg=NONE cterm=underline term=underline
hi SpellRare guifg=#9acd32 guibg=NONE guisp=#9acd32 gui=undercurl ctermfg=112 ctermbg=NONE cterm=underline term=underline
hi Statement guifg=#f0e68c guibg=NONE guisp=NONE gui=bold ctermfg=186 ctermbg=NONE cterm=bold term=NONE
hi StatusLine guifg=#333333 guibg=#c2bfa5 guisp=NONE gui=NONE ctermfg=236 ctermbg=144 cterm=NONE term=bold,reverse
hi StatusLineNC guifg=#7f7f8c guibg=#c2bfa5 guisp=NONE gui=NONE ctermfg=242 ctermbg=144 cterm=NONE term=bold,underline
hi StatusLineTerm guifg=#333333 guibg=#c2bfa5 guisp=NONE gui=NONE ctermfg=236 ctermbg=144 cterm=NONE term=bold,reverse
hi StatusLineTermNC guifg=#ffffff guibg=#c2bfa5 guisp=NONE gui=NONE ctermfg=231 ctermbg=144 cterm=NONE term=bold,underline
hi TabLine guifg=#333333 guibg=#c2bfa5 guisp=NONE gui=NONE ctermfg=236 ctermbg=144 cterm=NONE term=bold,underline
hi TabLineFill guifg=NONE guibg=#c2bfa5 guisp=NONE gui=NONE ctermfg=NONE ctermbg=144 cterm=NONE term=NONE
hi TabLineSel guifg=#333333 guibg=#f0e68c guisp=NONE gui=NONE ctermfg=236 ctermbg=186 cterm=NONE term=bold,reverse
hi Title guifg=#cd5c5c guibg=NONE guisp=NONE gui=bold ctermfg=167 ctermbg=NONE cterm=bold term=NONE
hi TitleBar guifg=#ffffff guibg=#4c4c4c guisp=NONE gui=NONE ctermfg=231 ctermbg=237 cterm=NONE term=NONE
hi TitleBarNC guifg=#8a7f7f guibg=#3c3c3c guisp=NONE gui=NONE ctermfg=244 ctermbg=236 cterm=NONE term=NONE
hi Todo guifg=#ff0000 guibg=#eeee00 guisp=NONE gui=NONE ctermfg=196 ctermbg=226 cterm=NONE term=bold,reverse
hi ToolbarButton guifg=#333333 guibg=#ffde9b guisp=NONE gui=bold ctermfg=236 ctermbg=222 cterm=bold term=bold,reverse
hi ToolbarLine guifg=NONE guibg=#666666 guisp=NONE gui=NONE ctermfg=NONE ctermbg=241 cterm=NONE term=reverse
hi Type guifg=#bdb76b guibg=NONE guisp=NONE gui=bold ctermfg=143 ctermbg=NONE cterm=bold term=NONE
hi Underlined guifg=#75a0ff guibg=NONE guisp=NONE gui=underline ctermfg=111 ctermbg=NONE cterm=underline term=underline
hi VertSplit guifg=#7f7f8c guibg=#c2bfa5 guisp=NONE gui=NONE ctermfg=242 ctermbg=144 cterm=NONE term=NONE
hi Visual guifg=#f0e68c guibg=#6b8e24 guisp=NONE gui=NONE ctermfg=186 ctermbg=64 cterm=NONE term=reverse
hi VisualNOS guifg=#f0e68c guibg=#6dceeb guisp=NONE gui=NONE ctermfg=186 ctermbg=81 cterm=NONE term=NONE
hi WarningMsg guifg=#cd5c5c guibg=NONE guisp=NONE gui=bold ctermfg=167 ctermbg=NONE cterm=bold term=standout
hi WildMenu guifg=#333333 guibg=#eeee00 guisp=NONE gui=NONE ctermfg=236 ctermbg=226 cterm=NONE term=bold
hi debugBreakpoint guifg=#ffa0a0 guibg=NONE guisp=NONE gui=reverse ctermfg=217 ctermbg=NONE cterm=reverse term=reverse
hi debugPC guifg=#666666 guibg=NONE guisp=NONE gui=reverse ctermfg=241 ctermbg=NONE cterm=reverse term=reverse

if has('gui_running') || s:tgc
  hi Cursor guifg=#333333 guibg=#f0e68c guisp=NONE gui=NONE cterm=NONE
  hi lCursor guifg=#333333 guibg=#ff0000 guisp=NONE gui=NONE cterm=NONE
endif

if s:tgc || s:t_Co >= 256
  finish
endif

if s:t_Co >= 16
  hi Normal ctermfg=white ctermbg=black cterm=NONE
  hi ColorColumn ctermfg=white ctermbg=darkred cterm=NONE
  hi Comment ctermfg=cyan ctermbg=NONE cterm=NONE
  hi Conceal ctermfg=grey ctermbg=NONE cterm=NONE
  hi Constant ctermfg=darkmagenta ctermbg=NONE cterm=NONE
  hi CursorColumn ctermfg=NONE ctermbg=NONE cterm=NONE
  hi CursorLine ctermfg=NONE ctermbg=NONE cterm=underline
  hi CursorLineNr ctermfg=darkyellow ctermbg=NONE cterm=bold
  hi DiffAdd ctermfg=white ctermbg=darkgreen cterm=NONE
  hi DiffChange ctermfg=white ctermbg=blue cterm=NONE
  hi DiffDelete ctermfg=white ctermbg=magenta cterm=NONE
  hi DiffText ctermfg=black ctermbg=grey cterm=NONE
  hi Directory ctermfg=blue ctermbg=NONE cterm=NONE
  hi Error ctermfg=red ctermbg=white cterm=reverse
  hi ErrorMsg ctermfg=red ctermbg=white cterm=reverse
  hi FoldColumn ctermfg=darkyellow ctermbg=darkgrey cterm=NONE
  hi Folded ctermfg=darkblue ctermbg=NONE cterm=NONE
  hi Identifier ctermfg=green ctermbg=NONE cterm=NONE
  hi Ignore ctermfg=NONE ctermbg=NONE cterm=NONE
  hi IncSearch ctermfg=black ctermbg=yellow cterm=NONE
  hi LineNr ctermfg=darkyellow ctermbg=NONE cterm=NONE
  hi MatchParen ctermfg=black ctermbg=darkyellow cterm=NONE
  hi ModeMsg ctermfg=magenta ctermbg=NONE cterm=bold
  hi MoreMsg ctermfg=darkgreen ctermbg=NONE cterm=bold
  hi NonText ctermfg=blue ctermbg=NONE cterm=NONE
  hi Pmenu ctermfg=NONE ctermbg=darkgrey cterm=NONE
  hi PmenuMatch ctermfg=NONE ctermbg=darkgrey cterm=bold
  hi PmenuMatchSel ctermfg=black ctermbg=yellow cterm=bold
  hi PmenuSbar ctermfg=NONE ctermbg=black cterm=NONE
  hi PmenuSel ctermfg=black ctermbg=yellow cterm=NONE
  hi PmenuThumb ctermfg=NONE ctermbg=white cterm=NONE
  hi PreProc ctermfg=darkred ctermbg=NONE cterm=NONE
  hi Question ctermfg=green ctermbg=NONE cterm=bold
  hi QuickFixLine ctermfg=black ctermbg=yellow cterm=NONE
  hi Search ctermfg=black ctermbg=darkgreen cterm=NONE
  hi SignColumn ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Special ctermfg=magenta ctermbg=NONE cterm=NONE
  hi SpecialKey ctermfg=darkgreen ctermbg=NONE cterm=NONE
  hi SpellBad ctermfg=darkred ctermbg=darkyellow cterm=reverse
  hi SpellCap ctermfg=darkblue ctermbg=grey cterm=reverse
  hi SpellLocal ctermfg=darkyellow ctermbg=NONE cterm=reverse
  hi SpellRare ctermfg=darkgreen ctermbg=NONE cterm=reverse
  hi Statement ctermfg=yellow ctermbg=NONE cterm=bold
  hi StatusLine ctermfg=black ctermbg=grey cterm=NONE
  hi StatusLineNC ctermfg=darkgrey ctermbg=grey cterm=NONE
  hi StatusLineTerm ctermfg=black ctermbg=grey cterm=NONE
  hi StatusLineTermNC ctermfg=darkgrey ctermbg=grey cterm=NONE
  hi TabLine ctermfg=black ctermbg=grey cterm=NONE
  hi TabLineFill ctermfg=NONE ctermbg=white cterm=NONE
  hi TabLineSel ctermfg=white ctermbg=black cterm=NONE
  hi Title ctermfg=darkred ctermbg=NONE cterm=bold
  hi TitleBar ctermfg=white ctermbg=black cterm=NONE
  hi TitleBarNC ctermfg=darkgrey ctermbg=black cterm=NONE
  hi Todo ctermfg=red ctermbg=darkmagenta cterm=NONE
  hi ToolbarButton ctermfg=black ctermbg=darkyellow cterm=NONE
  hi ToolbarLine ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Type ctermfg=darkyellow ctermbg=NONE cterm=bold
  hi Underlined ctermfg=darkblue ctermbg=NONE cterm=underline
  hi VertSplit ctermfg=darkgrey ctermbg=grey cterm=NONE
  hi Visual ctermfg=white ctermbg=darkgreen cterm=NONE
  hi VisualNOS ctermfg=NONE ctermbg=NONE cterm=underline
  hi WarningMsg ctermfg=darkred ctermbg=NONE cterm=bold
  hi WildMenu ctermfg=black ctermbg=darkmagenta cterm=NONE
  hi debugBreakpoint ctermfg=cyan ctermbg=NONE cterm=reverse
  hi debugPC ctermfg=grey ctermbg=NONE cterm=reverse
  finish
endif

if s:t_Co >= 8
  hi Normal ctermfg=grey ctermbg=black cterm=NONE
  hi ColorColumn ctermfg=grey ctermbg=darkred cterm=NONE
  hi Comment ctermfg=darkcyan ctermbg=NONE cterm=bold
  hi Conceal ctermfg=grey ctermbg=NONE cterm=NONE
  hi Constant ctermfg=darkmagenta ctermbg=NONE cterm=NONE
  hi CursorColumn ctermfg=NONE ctermbg=NONE cterm=NONE
  hi CursorLine ctermfg=NONE ctermbg=NONE cterm=underline
  hi CursorLineNr ctermfg=darkyellow ctermbg=NONE cterm=bold
  hi DiffAdd ctermfg=white ctermbg=darkgreen cterm=NONE
  hi DiffChange ctermfg=white ctermbg=darkblue cterm=NONE
  hi DiffDelete ctermfg=white ctermbg=darkmagenta cterm=NONE
  hi DiffText ctermfg=black ctermbg=grey cterm=NONE
  hi Directory ctermfg=darkblue ctermbg=NONE cterm=bold
  hi Error ctermfg=darkred ctermbg=grey cterm=reverse
  hi ErrorMsg ctermfg=darkred ctermbg=grey cterm=reverse
  hi FoldColumn ctermfg=darkyellow ctermbg=NONE cterm=NONE
  hi Folded ctermfg=darkblue ctermbg=NONE cterm=NONE
  hi Identifier ctermfg=darkgreen ctermbg=NONE cterm=NONE
  hi Ignore ctermfg=NONE ctermbg=NONE cterm=NONE
  hi IncSearch ctermfg=black ctermbg=darkyellow cterm=NONE
  hi LineNr ctermfg=darkyellow ctermbg=NONE cterm=NONE
  hi MatchParen ctermfg=black ctermbg=darkyellow cterm=NONE
  hi ModeMsg ctermfg=darkmagenta ctermbg=NONE cterm=bold
  hi MoreMsg ctermfg=darkgreen ctermbg=NONE cterm=bold
  hi NonText ctermfg=darkblue ctermbg=NONE cterm=bold
  hi Pmenu ctermfg=black ctermbg=darkcyan cterm=NONE
  hi PmenuMatch ctermfg=black ctermbg=darkcyan cterm=bold
  hi PmenuMatchSel ctermfg=black ctermbg=darkyellow cterm=bold
  hi PmenuSbar ctermfg=NONE ctermbg=black cterm=NONE
  hi PmenuSel ctermfg=black ctermbg=darkyellow cterm=NONE
  hi PmenuThumb ctermfg=NONE ctermbg=grey cterm=NONE
  hi PreProc ctermfg=darkred ctermbg=NONE cterm=NONE
  hi Question ctermfg=darkgreen ctermbg=NONE cterm=bold
  hi QuickFixLine ctermfg=black ctermbg=darkyellow cterm=NONE
  hi Search ctermfg=black ctermbg=darkgreen cterm=NONE
  hi SignColumn ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Special ctermfg=darkmagenta ctermbg=NONE cterm=bold
  hi SpecialKey ctermfg=darkgreen ctermbg=NONE cterm=NONE
  hi SpellBad ctermfg=darkred ctermbg=darkyellow cterm=reverse
  hi SpellCap ctermfg=darkblue ctermbg=grey cterm=reverse
  hi SpellLocal ctermfg=darkyellow ctermbg=NONE cterm=reverse
  hi SpellRare ctermfg=darkgreen ctermbg=NONE cterm=reverse
  hi Statement ctermfg=darkyellow ctermbg=NONE cterm=bold
  hi StatusLine ctermfg=grey ctermbg=black cterm=bold,reverse
  hi StatusLineNC ctermfg=grey ctermbg=black cterm=reverse
  hi StatusLineTerm ctermfg=grey ctermbg=black cterm=bold,reverse
  hi StatusLineTermNC ctermfg=grey ctermbg=black cterm=reverse
  hi TabLine ctermfg=black ctermbg=grey cterm=NONE
  hi TabLineFill ctermfg=NONE ctermbg=grey cterm=NONE
  hi TabLineSel ctermfg=grey ctermbg=black cterm=NONE
  hi Title ctermfg=darkred ctermbg=NONE cterm=bold
  hi TitleBar ctermfg=white ctermbg=black cterm=NONE
  hi TitleBarNC ctermfg=darkgrey ctermbg=black cterm=NONE
  hi Todo ctermfg=darkred ctermbg=darkmagenta cterm=NONE
  hi ToolbarButton ctermfg=grey ctermbg=black cterm=bold,reverse
  hi ToolbarLine ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Type ctermfg=darkyellow ctermbg=NONE cterm=bold
  hi Underlined ctermfg=NONE ctermbg=NONE cterm=underline
  hi VertSplit ctermfg=grey ctermbg=black cterm=reverse
  hi Visual ctermfg=NONE ctermbg=NONE cterm=reverse
  hi VisualNOS ctermfg=NONE ctermbg=NONE cterm=underline
  hi WarningMsg ctermfg=darkred ctermbg=NONE cterm=bold
  hi WildMenu ctermfg=black ctermbg=darkmagenta cterm=NONE
  hi debugBreakpoint ctermfg=darkcyan ctermbg=NONE cterm=reverse
  hi debugPC ctermfg=grey ctermbg=NONE cterm=reverse
  finish
endif

if s:t_Co >= 0
  hi CurSearch term=reverse
  hi Cursor term=reverse
  hi CursorIM term=NONE
  hi CursorLineFold term=underline
  hi CursorLineSign term=underline
  hi EndOfBuffer term=NONE
  hi Float term=NONE
  hi Function term=NONE
  hi Number term=NONE
  hi Terminal term=NONE
  finish
endif

" vim: et ts=8 sw=2 sts=2
