" Name:         delek
" Description:  Light background colorscheme.
" Author:       Original author David Schweikert <david@schweikert.ch>
" Maintainer:   Original maintainer David Schweikert <david@schweikert.ch>
" URL:          https://github.com/vim/colorschemes
" License:      Same as Vim
" Last Change:  2025 Oct 08

" Generated by Colortemplate v3.0.0-beta10

set background=light

" hi clear
source $VIMRUNTIME/colors/vim.lua " Nvim: revert to Vim default color scheme
let g:colors_name = 'delek'

let s:t_Co = &t_Co
let s:tgc = has('termguicolors') && &termguicolors

let g:terminal_ansi_colors = ['#ffffff', '#0000ff', '#00cd00', '#cd00cd', '#008b8b', '#0000ff', '#ff1493', '#bcbcbc', '#ee0000', '#0000ff', '#00cd00', '#cd00cd', '#008b8b', '#0000ff', '#ff1493', '#000000']
" Nvim uses g:terminal_color_{0-15} instead
for i in range(g:terminal_ansi_colors->len())
  let g:terminal_color_{i} = g:terminal_ansi_colors[i]
endfor

hi! link CurSearch Search
hi! link CursorIM Cursor
hi! link CursorLineFold CursorLine
hi! link CursorLineSign CursorLine
hi! link ErrorMsg Error
hi! link Float Number
hi! link Function Identifier
hi! link LineNrAbove LineNr
hi! link LineNrBelow LineNr
hi! link MessageWindow Pmenu
hi! link Number Constant
hi! link PopupNotification Todo
hi! link Terminal Normal

hi Normal guifg=#000000 guibg=#ffffff guisp=NONE gui=NONE ctermfg=16 ctermbg=231 cterm=NONE term=NONE
hi ColorColumn guifg=#000000 guibg=#e4e4e4 guisp=NONE gui=NONE ctermfg=16 ctermbg=254 cterm=NONE term=reverse
hi Comment guifg=#ee0000 guibg=NONE guisp=NONE gui=NONE ctermfg=196 ctermbg=NONE cterm=NONE term=bold
hi Conceal guifg=#bcbcbc guibg=NONE guisp=NONE gui=NONE ctermfg=250 ctermbg=NONE cterm=NONE term=NONE
hi Constant guifg=#00cd00 guibg=NONE guisp=NONE gui=NONE ctermfg=40 ctermbg=NONE cterm=NONE term=NONE
hi Cursor guifg=#ffffff guibg=#000000 guisp=NONE gui=NONE ctermfg=231 ctermbg=16 cterm=NONE term=reverse
hi CursorColumn guifg=NONE guibg=#e4e4e4 guisp=NONE gui=NONE ctermfg=NONE ctermbg=254 cterm=NONE term=NONE
hi CursorLine guifg=NONE guibg=#e4e4e4 guisp=NONE gui=NONE ctermfg=NONE ctermbg=254 cterm=NONE term=underline
hi CursorLineNr guifg=#a52a2a guibg=NONE guisp=NONE gui=bold ctermfg=124 ctermbg=NONE cterm=bold term=bold
hi DiffAdd guifg=#ffffff guibg=#5f875f guisp=NONE gui=NONE ctermfg=231 ctermbg=65 cterm=NONE term=reverse
hi DiffChange guifg=#ffffff guibg=#5f87af guisp=NONE gui=NONE ctermfg=231 ctermbg=67 cterm=NONE term=NONE
hi DiffDelete guifg=#ffffff guibg=#af5faf guisp=NONE gui=NONE ctermfg=231 ctermbg=133 cterm=NONE term=reverse
hi DiffText guifg=#000000 guibg=#c6c6c6 guisp=NONE gui=NONE ctermfg=16 ctermbg=251 cterm=NONE term=reverse
hi Directory guifg=#008b8b guibg=NONE guisp=NONE gui=bold ctermfg=30 ctermbg=NONE cterm=bold term=NONE
hi EndOfBuffer guifg=#bcbcbc guibg=NONE guisp=NONE gui=NONE ctermfg=250 ctermbg=NONE cterm=NONE term=NONE
hi Error guifg=#ff0000 guibg=#ffffff guisp=NONE gui=reverse ctermfg=196 ctermbg=231 cterm=reverse term=bold,reverse
hi FoldColumn guifg=#00008b guibg=NONE guisp=NONE gui=NONE ctermfg=18 ctermbg=NONE cterm=NONE term=NONE
hi Folded guifg=#00008b guibg=#e4e4e4 guisp=NONE gui=NONE ctermfg=18 ctermbg=254 cterm=NONE term=NONE
hi Identifier guifg=#008b8b guibg=NONE guisp=NONE gui=NONE ctermfg=30 ctermbg=NONE cterm=NONE term=NONE
hi Ignore guifg=NONE guibg=NONE guisp=NONE gui=NONE ctermfg=NONE ctermbg=NONE cterm=NONE term=NONE
hi IncSearch guifg=#00cd00 guibg=NONE guisp=NONE gui=reverse ctermfg=40 ctermbg=NONE cterm=reverse term=bold,reverse,underline
hi LineNr guifg=#a52a2a guibg=NONE guisp=NONE gui=NONE ctermfg=124 ctermbg=NONE cterm=NONE term=NONE
hi MatchParen guifg=#ffffff guibg=#ff1493 guisp=NONE gui=NONE ctermfg=231 ctermbg=198 cterm=NONE term=bold,underline
hi ModeMsg guifg=#000000 guibg=#ffffff guisp=NONE gui=bold ctermfg=16 ctermbg=231 cterm=bold term=bold
hi MoreMsg guifg=#000000 guibg=#ffffff guisp=NONE gui=bold ctermfg=16 ctermbg=231 cterm=bold term=NONE
hi NonText guifg=#bcbcbc guibg=NONE guisp=NONE gui=NONE ctermfg=250 ctermbg=NONE cterm=NONE term=NONE
hi Pmenu guifg=#000000 guibg=#add8e6 guisp=NONE gui=NONE ctermfg=16 ctermbg=152 cterm=NONE term=reverse
hi PmenuMatch guifg=#cd00cd guibg=#add8e6 guisp=NONE gui=NONE ctermfg=164 ctermbg=152 cterm=NONE term=NONE
hi PmenuMatchSel guifg=#ff87ff guibg=#00008b guisp=NONE gui=NONE ctermfg=213 ctermbg=18 cterm=NONE term=NONE
hi PmenuSbar guifg=NONE guibg=#ffffff guisp=NONE gui=NONE ctermfg=NONE ctermbg=231 cterm=NONE term=reverse
hi PmenuSel guifg=#ffffff guibg=#00008b guisp=NONE gui=NONE ctermfg=231 ctermbg=18 cterm=NONE term=bold
hi PmenuThumb guifg=NONE guibg=#008b8b guisp=NONE gui=NONE ctermfg=NONE ctermbg=30 cterm=NONE term=NONE
hi PreProc guifg=#cd00cd guibg=NONE guisp=NONE gui=NONE ctermfg=164 ctermbg=NONE cterm=NONE term=NONE
hi Question guifg=#008700 guibg=NONE guisp=NONE gui=bold ctermfg=28 ctermbg=NONE cterm=bold term=standout
hi QuickFixLine guifg=#ffffff guibg=#008b8b guisp=NONE gui=NONE ctermfg=231 ctermbg=30 cterm=NONE term=NONE
hi Search guifg=#ffffff guibg=#cd00cd guisp=NONE gui=NONE ctermfg=231 ctermbg=164 cterm=NONE term=reverse
hi SignColumn guifg=NONE guibg=NONE guisp=NONE gui=NONE ctermfg=NONE ctermbg=NONE cterm=NONE term=reverse
hi Special guifg=#ff1493 guibg=NONE guisp=NONE gui=NONE ctermfg=198 ctermbg=NONE cterm=NONE term=NONE
hi SpecialKey guifg=#bcbcbc guibg=NONE guisp=NONE gui=NONE ctermfg=250 ctermbg=NONE cterm=NONE term=bold
hi SpellBad guifg=#ff0000 guibg=NONE guisp=#ff0000 gui=undercurl ctermfg=196 ctermbg=NONE cterm=underline term=underline
hi SpellCap guifg=#008b8b guibg=NONE guisp=#008b8b gui=undercurl ctermfg=30 ctermbg=NONE cterm=underline term=underline
hi SpellLocal guifg=#0000ff guibg=NONE guisp=#0000ff gui=undercurl ctermfg=21 ctermbg=NONE cterm=underline term=underline
hi SpellRare guifg=#cd00cd guibg=NONE guisp=#cd00cd gui=undercurl ctermfg=164 ctermbg=NONE cterm=underline term=underline
hi Statement guifg=#0000ff guibg=NONE guisp=NONE gui=bold ctermfg=21 ctermbg=NONE cterm=bold term=NONE
hi StatusLine guifg=#ffff00 guibg=#00008b guisp=NONE gui=bold ctermfg=226 ctermbg=18 cterm=bold term=bold,reverse
hi StatusLineNC guifg=#ffd700 guibg=#00008b guisp=NONE gui=NONE ctermfg=220 ctermbg=18 cterm=NONE term=bold,underline
hi StatusLineTerm guifg=#ffff00 guibg=#0000ff guisp=NONE gui=bold ctermfg=226 ctermbg=21 cterm=bold term=bold,reverse
hi StatusLineTermNC guifg=#e4e4e4 guibg=#0000ff guisp=NONE gui=NONE ctermfg=254 ctermbg=21 cterm=NONE term=bold,underline
hi TabLine guifg=#000000 guibg=#e4e4e4 guisp=NONE gui=NONE ctermfg=16 ctermbg=254 cterm=NONE term=bold,underline
hi TabLineFill guifg=NONE guibg=#bcbcbc guisp=NONE gui=NONE ctermfg=NONE ctermbg=250 cterm=NONE term=NONE
hi TabLineSel guifg=#000000 guibg=#ffffff guisp=NONE gui=bold ctermfg=16 ctermbg=231 cterm=bold term=bold,reverse
hi Title guifg=#cd00cd guibg=NONE guisp=NONE gui=bold ctermfg=164 ctermbg=NONE cterm=bold term=NONE
hi TitleBar guifg=#000000 guibg=#ececec guisp=NONE gui=NONE ctermfg=16 ctermbg=255 cterm=NONE term=NONE
hi TitleBarNC guifg=#878787 guibg=#f5f5f5 guisp=NONE gui=NONE ctermfg=102 ctermbg=255 cterm=NONE term=NONE
hi Todo guifg=#000000 guibg=#ffff00 guisp=NONE gui=NONE ctermfg=16 ctermbg=226 cterm=NONE term=bold,reverse
hi ToolbarButton guifg=#ffffff guibg=#bcbcbc guisp=NONE gui=bold ctermfg=231 ctermbg=250 cterm=bold term=bold,reverse
hi ToolbarLine guifg=NONE guibg=#e4e4e4 guisp=NONE gui=NONE ctermfg=NONE ctermbg=254 cterm=NONE term=reverse
hi Type guifg=#0000ff guibg=NONE guisp=NONE gui=bold ctermfg=21 ctermbg=NONE cterm=bold term=NONE
hi Underlined guifg=#6a5acd guibg=NONE guisp=NONE gui=underline ctermfg=62 ctermbg=NONE cterm=underline term=underline
hi VertSplit guifg=#e4e4e4 guibg=#00008b guisp=NONE gui=NONE ctermfg=254 ctermbg=18 cterm=NONE term=NONE
hi Visual guifg=#000000 guibg=#d0d0d0 guisp=NONE gui=NONE ctermfg=16 ctermbg=252 cterm=NONE term=reverse
hi VisualNOS guifg=NONE guibg=#ee0000 guisp=NONE gui=NONE ctermfg=NONE ctermbg=196 cterm=NONE term=NONE
hi WarningMsg guifg=#cd00cd guibg=#ffffff guisp=NONE gui=NONE ctermfg=164 ctermbg=231 cterm=NONE term=standout
hi WildMenu guifg=#00008b guibg=#ffd700 guisp=NONE gui=bold ctermfg=18 ctermbg=220 cterm=bold term=bold
hi debugBreakpoint guifg=#ffff00 guibg=#0000ff guisp=NONE gui=NONE ctermfg=226 ctermbg=21 cterm=NONE term=NONE
hi debugPC guifg=#ff1493 guibg=#0000ff guisp=NONE gui=NONE ctermfg=198 ctermbg=21 cterm=NONE term=NONE
hi lCursor guifg=#000000 guibg=#00cd00 guisp=NONE gui=NONE ctermfg=16 ctermbg=40 cterm=NONE term=NONE

if s:tgc || s:t_Co >= 256
  finish
endif

if s:t_Co >= 16
  hi Normal ctermfg=black ctermbg=white cterm=NONE
  hi ColorColumn ctermfg=black ctermbg=grey cterm=NONE
  hi Comment ctermfg=darkred ctermbg=NONE cterm=NONE
  hi Conceal ctermfg=darkgrey ctermbg=NONE cterm=NONE
  hi Constant ctermfg=darkgreen ctermbg=NONE cterm=NONE
  hi Cursor ctermfg=white ctermbg=black cterm=NONE
  hi CursorColumn ctermfg=NONE ctermbg=grey cterm=NONE
  hi CursorLine ctermfg=NONE ctermbg=NONE cterm=underline
  hi CursorLineNr ctermfg=darkgrey ctermbg=NONE cterm=bold
  hi DiffAdd ctermfg=white ctermbg=darkgreen cterm=NONE
  hi DiffChange ctermfg=white ctermbg=blue cterm=NONE
  hi DiffDelete ctermfg=white ctermbg=magenta cterm=NONE
  hi DiffText ctermfg=black ctermbg=grey cterm=NONE
  hi Directory ctermfg=darkcyan ctermbg=NONE cterm=bold
  hi EndOfBuffer ctermfg=darkgrey ctermbg=NONE cterm=NONE
  hi Error ctermfg=red ctermbg=white cterm=reverse
  hi FoldColumn ctermfg=darkblue ctermbg=NONE cterm=NONE
  hi Folded ctermfg=darkblue ctermbg=grey cterm=NONE
  hi Identifier ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Ignore ctermfg=NONE ctermbg=NONE cterm=NONE
  hi IncSearch ctermfg=darkgreen ctermbg=NONE cterm=reverse
  hi LineNr ctermfg=darkgrey ctermbg=NONE cterm=NONE
  hi MatchParen ctermfg=white ctermbg=magenta cterm=NONE
  hi ModeMsg ctermfg=black ctermbg=white cterm=bold
  hi MoreMsg ctermfg=black ctermbg=white cterm=bold
  hi NonText ctermfg=darkgrey ctermbg=NONE cterm=NONE
  hi Pmenu ctermfg=black ctermbg=grey cterm=NONE
  hi PmenuMatch ctermfg=darkmagenta ctermbg=grey cterm=NONE
  hi PmenuMatchSel ctermfg=magenta ctermbg=darkblue cterm=NONE
  hi PmenuSbar ctermfg=NONE ctermbg=white cterm=NONE
  hi PmenuSel ctermfg=white ctermbg=darkblue cterm=NONE
  hi PmenuThumb ctermfg=NONE ctermbg=darkcyan cterm=NONE
  hi PreProc ctermfg=darkmagenta ctermbg=NONE cterm=NONE
  hi Question ctermfg=darkgreen ctermbg=NONE cterm=bold
  hi QuickFixLine ctermfg=white ctermbg=darkcyan cterm=NONE
  hi Search ctermfg=white ctermbg=darkmagenta cterm=NONE
  hi SignColumn ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Special ctermfg=magenta ctermbg=NONE cterm=NONE
  hi SpecialKey ctermfg=darkgrey ctermbg=NONE cterm=NONE
  hi SpellBad ctermfg=red ctermbg=NONE cterm=underline
  hi SpellCap ctermfg=darkcyan ctermbg=NONE cterm=underline
  hi SpellLocal ctermfg=darkblue ctermbg=NONE cterm=underline
  hi SpellRare ctermfg=darkmagenta ctermbg=NONE cterm=underline
  hi Statement ctermfg=darkblue ctermbg=NONE cterm=bold
  hi StatusLine ctermfg=yellow ctermbg=darkblue cterm=bold
  hi StatusLineNC ctermfg=darkyellow ctermbg=darkblue cterm=NONE
  hi StatusLineTerm ctermfg=yellow ctermbg=blue cterm=bold
  hi StatusLineTermNC ctermfg=grey ctermbg=blue cterm=NONE
  hi TabLine ctermfg=black ctermbg=grey cterm=NONE
  hi TabLineFill ctermfg=NONE ctermbg=darkgrey cterm=NONE
  hi TabLineSel ctermfg=black ctermbg=white cterm=bold
  hi Title ctermfg=darkmagenta ctermbg=NONE cterm=bold
  hi TitleBar ctermfg=black ctermbg=white cterm=NONE
  hi TitleBarNC ctermfg=gray ctermbg=white cterm=NONE
  hi Todo ctermfg=black ctermbg=yellow cterm=NONE
  hi ToolbarButton ctermfg=white ctermbg=darkgrey cterm=bold
  hi ToolbarLine ctermfg=NONE ctermbg=grey cterm=NONE
  hi Type ctermfg=blue ctermbg=NONE cterm=bold
  hi Underlined ctermfg=blue ctermbg=NONE cterm=underline
  hi VertSplit ctermfg=grey ctermbg=darkblue cterm=NONE
  hi Visual ctermfg=white ctermbg=darkgrey cterm=NONE
  hi VisualNOS ctermfg=NONE ctermbg=darkred cterm=NONE
  hi WarningMsg ctermfg=darkmagenta ctermbg=white cterm=NONE
  hi WildMenu ctermfg=darkblue ctermbg=darkyellow cterm=bold
  hi debugBreakpoint ctermfg=yellow ctermbg=darkblue cterm=NONE
  hi debugPC ctermfg=magenta ctermbg=darkblue cterm=NONE
  hi lCursor ctermfg=black ctermbg=darkgreen cterm=NONE
  finish
endif

if s:t_Co >= 8
  hi Normal ctermfg=black ctermbg=gray cterm=NONE
  hi ColorColumn ctermfg=black ctermbg=darkred cterm=NONE
  hi Comment ctermfg=darkred ctermbg=NONE cterm=NONE
  hi Conceal ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Constant ctermfg=darkgreen ctermbg=NONE cterm=NONE
  hi Cursor ctermfg=white ctermbg=black cterm=NONE
  hi CursorColumn ctermfg=NONE ctermbg=NONE cterm=NONE
  hi CursorLine ctermfg=NONE ctermbg=grey cterm=NONE
  hi CursorLineNr ctermfg=NONE ctermbg=NONE cterm=bold
  hi DiffAdd ctermfg=white ctermbg=darkgreen cterm=NONE
  hi DiffChange ctermfg=white ctermbg=darkblue cterm=NONE
  hi DiffDelete ctermfg=white ctermbg=darkmagenta cterm=NONE
  hi DiffText ctermfg=black ctermbg=grey cterm=NONE
  hi Directory ctermfg=darkcyan ctermbg=NONE cterm=bold
  hi EndOfBuffer ctermfg=darkblue ctermbg=NONE cterm=NONE
  hi Error ctermfg=red ctermbg=gray cterm=reverse
  hi FoldColumn ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Folded ctermfg=darkyellow ctermbg=NONE cterm=NONE
  hi Identifier ctermfg=darkcyan ctermbg=NONE cterm=NONE
  hi Ignore ctermfg=NONE ctermbg=NONE cterm=NONE
  hi IncSearch ctermfg=darkgreen ctermbg=NONE cterm=reverse
  hi LineNr ctermfg=NONE ctermbg=NONE cterm=bold
  hi MatchParen ctermfg=black ctermbg=darkmagenta cterm=bold
  hi ModeMsg ctermfg=NONE ctermbg=NONE cterm=bold
  hi MoreMsg ctermfg=NONE ctermbg=NONE cterm=bold
  hi NonText ctermfg=darkblue ctermbg=NONE cterm=NONE
  hi Pmenu ctermfg=black ctermbg=darkcyan cterm=NONE
  hi PmenuMatch ctermfg=black ctermbg=darkcyan cterm=bold
  hi PmenuMatchSel ctermfg=black ctermbg=darkyellow cterm=bold
  hi PmenuSbar ctermfg=NONE ctermbg=NONE cterm=NONE
  hi PmenuSel ctermfg=black ctermbg=darkyellow cterm=NONE
  hi PmenuThumb ctermfg=NONE ctermbg=darkgreen cterm=NONE
  hi PreProc ctermfg=darkmagenta ctermbg=NONE cterm=NONE
  hi Question ctermfg=darkgreen ctermbg=NONE cterm=bold
  hi QuickFixLine ctermfg=black ctermbg=darkcyan cterm=NONE
  hi Search ctermfg=black ctermbg=darkmagenta cterm=NONE
  hi SignColumn ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Special ctermfg=darkmagenta ctermbg=NONE cterm=bold
  hi SpecialKey ctermfg=darkblue ctermbg=NONE cterm=NONE
  hi SpellBad ctermfg=darkred ctermbg=NONE cterm=reverse
  hi SpellCap ctermfg=darkcyan ctermbg=black cterm=reverse
  hi SpellLocal ctermfg=darkblue ctermbg=NONE cterm=reverse
  hi SpellRare ctermfg=darkmagenta ctermbg=NONE cterm=reverse
  hi Statement ctermfg=darkblue ctermbg=NONE cterm=bold
  hi StatusLine ctermfg=darkblue ctermbg=darkyellow cterm=bold,reverse
  hi StatusLineNC ctermfg=darkblue ctermbg=darkyellow cterm=reverse
  hi StatusLineTerm ctermfg=darkblue ctermbg=darkyellow cterm=bold,reverse
  hi StatusLineTermNC ctermfg=darkblue ctermbg=darkyellow cterm=reverse
  hi TabLine ctermfg=gray ctermbg=black cterm=NONE
  hi TabLineFill ctermfg=NONE ctermbg=black cterm=NONE
  hi TabLineSel ctermfg=black ctermbg=gray cterm=NONE
  hi Title ctermfg=darkmagenta ctermbg=NONE cterm=bold
  hi TitleBar ctermfg=black ctermbg=white cterm=NONE
  hi TitleBarNC ctermfg=gray ctermbg=white cterm=NONE
  hi Todo ctermfg=black ctermbg=darkyellow cterm=NONE
  hi ToolbarButton ctermfg=black ctermbg=gray cterm=bold,reverse
  hi ToolbarLine ctermfg=NONE ctermbg=NONE cterm=NONE
  hi Type ctermfg=darkblue ctermbg=NONE cterm=NONE
  hi Underlined ctermfg=darkblue ctermbg=NONE cterm=underline
  hi VertSplit ctermfg=darkblue ctermbg=darkyellow cterm=reverse
  hi Visual ctermfg=NONE ctermbg=NONE cterm=reverse
  hi VisualNOS ctermfg=NONE ctermbg=NONE cterm=underline
  hi WarningMsg ctermfg=darkmagenta ctermbg=NONE cterm=NONE
  hi WildMenu ctermfg=black ctermbg=darkyellow cterm=NONE
  hi debugBreakpoint ctermfg=darkyellow ctermbg=darkblue cterm=NONE
  hi debugPC ctermfg=darkmagenta ctermbg=darkblue cterm=NONE
  hi lCursor ctermfg=black ctermbg=darkgreen cterm=NONE
  finish
endif

if s:t_Co >= 0
  hi CurSearch term=reverse
  hi CursorIM term=NONE
  hi CursorLineFold term=underline
  hi CursorLineSign term=underline
  hi ErrorMsg term=bold,reverse
  hi Float term=NONE
  hi Function term=NONE
  hi Number term=NONE
  hi Terminal term=NONE
  finish
endif

" vim: et ts=8 sw=2 sts=2
